package edu.umich.popcup.util;

import java.io.Serializable;
import java.util.Vector;

import edu.umich.popcup.util.Image.Category;
import edu.umich.popcup.util.Image.Pose;

/**
 * Model3D is the object to transfer information from the server to mobile client.
 * To save bandwidth, only the vertices and faces are kept.
 * 
 * @author Fisher
 */
public class Model3D implements Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	
	private Vector<float[]> vertices;
	private Vector<int[]> colors;
	private Vector<short[]> indices;
	private Vector<float[]> texCoords;
	private Vector<byte[]> texBytes;
	private Pose pose;
	private Category cat;
	
	public Model3D() {
		vertices = new Vector<float[]>();
		colors = new Vector<int[]>();
		indices = new Vector<short[]>();
		texCoords = new Vector<float[]>();
		texBytes = new Vector<byte[]>();
	}
	
	public void setPose(Pose pose) {
		this.pose = pose;
	}
	
	public void setCategory(Category cat) {
		this.cat = cat;
	}
	
	public Pose getPose() {
		return pose;
	}
	
	public Category getCategory() {
		return cat;
	}
	
	public void addPly(Ply ply) {
		vertices.add(ply.getCoordArray());
		colors.add(ply.getColoFixedArray());
		indices.add(ply.getIndexArray());
		texCoords.add(ply.getTexCoordArray());
		texBytes.add(ply.getTexture());		
	}
	
	/**
	 * Check the number of the available ply models
	 * @return the number of ply models
	 */
	public int availPly() {
		return vertices.size();
	}
	
	public float[] getVertices(int idx) {
		return vertices.get(idx);
	}
	
	public int[] getColors(int idx) {
		return colors.get(idx);
	}
	
	public short[] getIndices(int idx) {
		return indices.get(idx);
	}
	
	public float[] getTexCoords(int idx) {
		return texCoords.get(idx);
	}
	
	public byte [] getTexture(int idx) {
		return texBytes.get(idx);
	}
}
