package edu.umich.popcup.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Serializable;
import java.io.StringWriter;
import java.io.Writer;
import java.util.Vector;

import android.R;
import android.os.Environment;

@SuppressWarnings("unused")
public class Ply implements Serializable {
	
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	public static void main (String [] args) throws IOException {
		System.out.println(new File(".").getCanonicalPath());
		System.out.println(System.getProperty("user.dir"));
		
		Ply ply = new Ply("./ply_res/mouse_1.ply");
		ply.readFile();
		
		ply.writeData(System.out);
	}
	
	enum ScalarType {
		INT8, UINT8, INT16, UINT16, INT32, UINT32, FLOAT32, FLOAT64, CHAR, UCHAR, SHORT, USHORT, INT, UINT, FLOAT, DOUBLE
	}
	
	/**
	 * 
	 * @author Fisher
	 *
	 *PropType
	 *	Type of the property
	 */
	enum PropType {
		SCALAR, LIST, STRING
	}
	
	/**
	 * @author Fisher
	 *
	 * The type of the input ply file. Currently, only ASCII is supported
	 */
	enum Format {
		ASCII, BINARY_BIG_ENDIAN, BINARY_SMALL_ENDIAN
	}
	
	static public class Property implements Serializable {

		private static final long serialVersionUID = 1L;
		
		public String name;
		public ScalarType countType;
		public ScalarType dataType;
		public PropType propType;
		public Number [][] data;
		public Number count;
		public int elemNum;
		public int dataIdx;
		
		public Property () {
			name = null;
			data = null;
			count = null;
			elemNum = 0;
			dataIdx = 0;
		}
	}
	
	public class Element implements Serializable {

		private static final long serialVersionUID = 1L;
		
		public String name;
		public int number;
		public int nprop;
		public Vector<Property> props;
		
		public Element() {
			number = -1;
			nprop = 0;
			props = new Vector<Property>();
		}
	}
	
	public Ply(String fileName) {
		reset();
		file = new File(fileName);
	}
	
	public Ply() {
		reset();
	}
	
	public Ply readFile(String fileName) {
		reset();
		file = new File(fileName);
		return readFile();
	}
	
	static public class Color implements Serializable {

		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;
		
		//Values are expressed as fixed point value internally
		//Its range is [0, 1)
		private int [] values;
		
		public Color() {
			values = new int[4];
			for (int i = 0; i < 4; i++) {
				values[i] = (1 << 16);
			}
		}
		
		// The normal set method assumes the r is uchar
		public void setR(int r) {
			values[0] = (r << 8);
		}
		
		public void setG(int g) {
			values[1] = (g << 8);
		}
		
		public void setB(int b) {
			values[2] = (b << 8);
		}
		
		public void setA(int a) {
			values[3] = (a << 8);
		}
		
		public int r() {
			return (values[0] >> 8);
		}
		
		public int g() {
			return (values[1] >> 8);
		}
		
		public int b() {
			return (values[2] >> 8);
		}
		
		public int a() {
			return (values[3] >> 8);
		}
		
		public int[] getFixedArray() {
			return values;
		}
		
		public int[] getUCharArray() {
			int [] result = new int[4];
			
			for (int i = 0; i < 4; i++) {
				result[i] = (values[i] >> 8);
			}
			
			return result;
		}
		
		public String toString() {
			return "r: " + r() + " g: " + g() + " b: " + b() + " a: " + a();
		}
	}
	
	static public class Coord implements Serializable {

		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;
		
		private float [] values;
		
		Coord() {
			values = new float[3];
		}
		
		public void setX(float x) {
			values[0] = x;
		}
		
		public void setY(float y) {
			values[1] = y;
		}
		
		public void setZ(float z) {
			values[2] = z;
		}
		
		public float x() {
			return values[0];
		}
		
		public float y() {
			return values[1];
		}
		
		public float z() {
			return values[2];
		}
		
		public float [] getArray() {
			return values;
		}
		
		public String toString() {
			return "x: " + x() + " y: " + y() + " z: " + z();
		}
	}
	
	static public class Vertex implements Serializable {

		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;
		
		public Coord coord;
		public Color color;
		
		public Vertex() {
			coord = null;
			color = null;
		}
		
		public void setProp(String name, Number value) {
			if (name.matches("[xyz]") && coord == null) {
				coord = new Coord();
			} else if (name.matches("red|green|blue|alpha") && color == null) {
				color = new Color();
			}
			if (name.equals("x")) {
				coord.setX(value.floatValue());
			} else if (name.equals("y")) {
				coord.setY(value.floatValue());
			} else if (name.equals("z")) {
				coord.setZ(value.floatValue());
			} else if (name.equals("red")) {
				color.setR(value.intValue());
			} else if (name.equals("green")) {
				color.setG(value.intValue());
			} else if (name.equals("blue")) {
				color.setB(value.intValue());
			} else if (name.equals("alpha")) {
				color.setA(value.intValue());
			} else {
				Log.e(TAG, name + " is not predefined property in vertex");
			}
		}
		
		public String toString() {
			String s;
			if (coord == null) {
				s = "Null vertex";
			} else {
				s = coord.toString();
				if (color == null) {
					s += " No color";
				} else {
					s += " " + color;
				}
			}
			return s;
		}
	}
	
	static public class Face implements Serializable {

		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;
		
		public int [] vtxIdx;
		public float [] texCoord;
		
		public Face() {
			vtxIdx = null;
			texCoord = null;
		}
		
		public void setList(String name, Number [] data) {
			if (name.equals("vertex_indices")) {
				vtxIdx = new int[data.length];
				for (int i = 0; i < data.length; i++)
					vtxIdx[i] = data[i].intValue();
			} else if (name.equals("texcoord")) {
				texCoord = new float[data.length];
				for (int i = 0; i < data.length; i++)
					texCoord[i] = data[i].floatValue();
			} else {
				Log.e(TAG, name+" is not predefined in face");
			}
		}
	}
	
	
	/**
	 * Member fields start from here
	 */
	
	
	public static final String TAG = "PLY";
	
	private Vector<Element> elems; // Elements
	private Format format;
	private float version; // Version number of the file
	private int elemIndex; // The current element being read
	private File file; // .ply file
	
	private String texName;
	private byte [] texBytes;
	
	private Vertex [] vertices;
	private Face [] faces;
	
	
	
	public Vertex[] getVertices() {
		return vertices;
	}
	
	public int[] getColoFixedArray() {
		int [] colorArr = new int[vertices.length * 4];
		for (int i = 0; i < vertices.length; i++) {
			System.arraycopy(vertices[i].color.getFixedArray(), 0, colorArr, i*4, 4);
		}
		return colorArr;
	}
	
	public float[] getCoordArray() {
		float [] coordArr = new float[vertices.length * 3];
		for (int i = 0; i < vertices.length; i++) {
			System.arraycopy(vertices[i].coord.getArray(), 0, coordArr, i*3, 3);
		}
		return coordArr;
	}
	
	/**
	 * The index is defined to be for the triangles 
	 * @return the indices of the vertices of the triangels
	 */
	public short[] getIndexArray() {
		short [] indices = new short[faces.length * 3];
		/*for (int i = 0; i < faces.length; i++) {
			System.arraycopy(faces[i].vtxIdx, 0, indices, i*3, 3);
		}*/
		for (int i = 0; i < faces.length; i++) {
			for (int j = 0; j < 3; j++) {
				indices[i*3+j] = (short)(faces[i].vtxIdx[j] & 0xFFFF);
			}
		}
		return indices;
	}
	
	public float[] getTexCoordArray() {
		float [] texCoords = new float[faces.length * 6];
		for (int i = 0; i < faces.length; i++) {
			System.arraycopy(faces[i].texCoord, 0, texCoords, i*6, 6);
		}
		return texCoords;
	}
	
	public byte[] getTexture() {
		return texBytes;
	}
	
	public Face[] getFaces() {
		return faces;
	}
	
	public static void writeVertices(Vertex [] vertices, Writer writer) throws IOException {
		for (Vertex v : vertices) {
			writer.write(v.toString() + "\n");
		}
	}
	
	
	/**
	 * 
	 * Parse the .ply file
	 * 
	 * @return this object on successful read and null otherwise
	 */
	public Ply readFile() {
		BufferedReader reader;
		try {
			reader = new BufferedReader(new InputStreamReader(new FileInputStream(file)));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			return null;
		}
		
		if (!readHeader(reader))
			return null;
		if (!readData(reader))
			return null;
		if (!readTexture())
			return null;

		return this;
	}
	
	public Ply readfile(InputStream in) {
		BufferedReader reader = new BufferedReader(new InputStreamReader(in));
		
		if (!readHeader(reader))
			return null;
		if (!readData(reader))
			return null;
		if (!readTexture())
			return null;
		
		return this;
	}
	
	private void addElement(String [] words) {
		Element elem = new Element();
		elem.name = words[1];
		elem.number = Integer.valueOf(words[2]).intValue();
		elem.nprop = 0;
		
		if (elem.name.equals("vertex")) {
			vertices = new Vertex[elem.number];
			for (int i = 0; i < vertices.length; i++)
				vertices[i] = new Vertex();
		}
		else if (elem.name.equals("face")) {
			faces = new Face[elem.number];
			for (int i = 0; i < faces.length; i++) 
				faces[i] = new Face();
		}
		elems.add(elem);
	}
	
	private void addProperty(String [] words) {
		Property prop = new Property();
		Element elem = elems.lastElement();

		if (words[1].equals("list")) {
			prop.countType = getPropType(words[2]);
			prop.dataType = getPropType(words[3]);
			prop.name = words[4];
			prop.propType = PropType.LIST;
		} else if (words[1].equals("string")) {
			prop.countType = ScalarType.INT8;
			prop.dataType = ScalarType.INT8;
			prop.name = words[2];
			prop.propType = PropType.STRING;
		} else {
			prop.countType = ScalarType.INT;
			prop.dataType = getPropType(words[1]);
			prop.name = words[2];
			prop.propType = PropType.SCALAR;
		}
		
		//Add this property to the current element
		prop.elemNum = elem.number;
		elem.props.add(prop);
		elem.nprop++;
		elems.set(elems.size()-1, elem);
	}
	
	private void addComment(String [] words) {
		if (words[1].equals("TextureFile")) {
			texName = words[2];
		}
	}
	
	private void addObjInfo(String [] words) {
		
	}
	
	public void writeData(String dataString) throws IOException {
		StringWriter writer = new StringWriter();
		writeData(writer);
		dataString = writer.toString();
	}
	
	public void writeData(OutputStream out) throws IOException {
		OutputStreamWriter writer = new OutputStreamWriter(out);
		writeData(writer);
	}
	
	public void writeData(Writer writer) throws IOException {
		for (Element elem: elems) {
			for (Property prop: elem.props) {
				prop.dataIdx = 0;
			}
			for (int i = 0; i < elem.number; i++) {
				for (Property prop: elem.props) {
					switch (prop.propType) {
					case LIST:
						writeOneList(writer, prop);
						break;
					case SCALAR:
						writeOneScalar(writer, prop);
						break;
					}
				}
				writer.write('\n');
			}
		}
	}
	
	private void writeOneList(Writer writer, Property prop) throws IOException {
		//writer.write(prop.count.toString());
		writeNumber(writer, prop.count, prop.countType);
		writer.write(" ");
		for (int i = 0; i < prop.count.intValue(); i++) {
			//writer.write(prop.data[prop.dataIdx][i].toString());
			writeNumber(writer, prop.data[prop.dataIdx][i], prop.dataType);
			writer.write(" ");
		}
		++prop.dataIdx;
	}
	
	private void writeOneScalar(Writer writer, Property prop) throws IOException {
		//writer.write(prop.data[prop.dataIdx++][0].toString());
		writeNumber(writer, prop.data[prop.dataIdx++][0], prop.dataType);
		writer.write(" ");
	}
	
	private void writeNumber(Writer writer, Number num, ScalarType type) throws IOException {
		switch (type) {
		//INT8, UINT8, INT16, UINT16, INT32, UINT32, FLOAT32, FLOAT64, CHAR, UCHAR, SHORT, USHORT, INT, UINT, FLOAT, DOUBLE
		case INT8: // Sometimes, I did this kind of silly thing......
		case UINT8:
		case INT16:
		case UINT16:
		case INT32:
		case UINT32:
		case CHAR:
		case UCHAR:
		case SHORT:
		case USHORT:
		case INT:
		case UINT:
			writer.write(Integer.toString(num.intValue()));
			break;
		default:
			writer.write(num.toString());
		}
	}
	
	/**
	 * Reader the header of the .ply file
	 * 
	 * @param reader
	 * @return true on successfully read and false otherwise
	 */
	private boolean readHeader(BufferedReader reader) {
		try {
			String line = reader.readLine();
			
			while (line != null) {
				String [] words = line.split(" ");
				/* parse words */
				if (words[0].equals("format")) {
					if (words.length != 3)
						return false;
					
					if (words[1].equals("ascii")) {
						format = Format.ASCII;
					} else if (words[1].equals("binary_big_endian")) {
						format = Format.BINARY_BIG_ENDIAN;
					} else if (words[1].equals("binary_small_endian")) {
						format = Format.BINARY_SMALL_ENDIAN;
					} else {
						return false;
					}
					version = Float.valueOf(words[2]).floatValue();
				} else if (words[0].equals("element")) {
					addElement(words);
				} else if (words[0].equals("property")) {
					addProperty(words);
				} else if (words[0].equals("comment")) {
					addComment(words);
				} else if (words[0].equals("obj_info")) {
					addObjInfo(words);
				} else if (words[0].equals("end_header"))
					break;
				
				line = reader.readLine();
			}			
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		return true;
	}
	
	private boolean readData(BufferedReader reader) {
		try {
			int lineIdx = 0;
			for (Element elem: elems) {
				for (int i = 0; i < elem.number; i++) {
					String line = reader.readLine();
					if (line == null) {
						Log.e(TAG, "Not enough data lines");
						return false;
					} else
						lineIdx++;
					
					String [] words = line.split(" ");
					int curIdx = 0;
					for (Property prop: elem.props) {
						switch (prop.propType) {
						case LIST:
							curIdx += readList(elem, prop, words, curIdx);
							break;
						case SCALAR:
							curIdx += readScalar(elem, prop, words, curIdx);
							break;
						case STRING:
							Log.e(TAG, "Unimplemented property");
							return false;
						}
					}
					
					if (curIdx != words.length) {
						Log.e(TAG, "There are too many words on Line " + lineIdx);
						return false;
					}
				}
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
		return true;
	}
	
	private boolean readTexture() {
		if (texName == null) {
			return true;
		} else {
			try {
				// For test purpose, I use the project specific texture name
				//String texPath = Environment.getExternalStorageDirectory().getCanonicalPath() + "/" + texName;
				//System.out.println(texPath);
				//FileInputStream fis = new FileInputStream(texPath);
				//String texPath = "/res/ply_res/"+texName;
				String texPath = "/Users/Fisher/code/eclipse/.metadata/.plugins/org.eclipse.wst.server.core/tmp0/wtpwebapps/ServerTest/res/ply_res/" + texName;
				FileInputStream fis = new FileInputStream(texPath);
				texBytes = new byte[fis.available()];
				int offset = 0;
				int len = texBytes.length;
				while (len != 0) {
					int numBytes = fis.read(texBytes, offset, len);
					offset += numBytes;
					len -= numBytes;
				}
			} catch (FileNotFoundException e) {
				e.printStackTrace();
				return false;
			} catch (IOException e) {
				e.printStackTrace();
				return false;
			}
		}
		
		return true;
	}
	
	/**
	 * Read a list of values.
	 * @param words
	 * @param from The beginning index
	 * @return the number of words read
	 */
	private int readList(Element elem, Property prop, String [] words, int from) {
		if (prop.count == null) {
			prop.count = Double.valueOf(words[from]);
			prop.data = new Number[prop.elemNum][prop.count.intValue()];
			prop.dataIdx = 0;
		} else {
			if (!prop.count.equals(Double.valueOf(words[from]))) {
				Log.e(TAG, "The numbers of property in a list doesn't match");
			}
		}
		
		for (int i = 0; i < prop.count.intValue(); i++) {
			prop.data[prop.dataIdx][i] = Double.valueOf(words[from+1+i]);
		}
		
		if (elem.name.equals("face")) {
			faces[prop.dataIdx].setList(prop.name, prop.data[prop.dataIdx]);
		}
		
		++prop.dataIdx;
		
		return prop.count.intValue() + 1;
	}
	
	/**
	 * Read a scalar from the array
	 * @param words the String array contains the data
	 * @param from the beginning index
	 * @return number of words read
	 */
	private int readScalar(Element elem, Property prop, String [] words, int from) {
		if (prop.count == null) {
			prop.count = new Integer(1);
			prop.data = new Number[prop.elemNum][1];
			prop.dataIdx = 0;
		}
		
		prop.data[prop.dataIdx][0] = Double.valueOf(words[from]);
		
		if (elem.name.equals("vertex")) {
			try {
				vertices[prop.dataIdx].setProp(prop.name, prop.data[prop.dataIdx][0]);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		
		++prop.dataIdx;
		
		return 1;
	}
	
	private ScalarType getPropType(String type) {
		if (type.equals("int8"))
			return ScalarType.INT8;
		else if (type.equals("uint8")) 
			return ScalarType.UINT8;
		else if (type.equals("int16")) 
			return ScalarType.INT16;
		else if (type.equals("uint16"))
			return ScalarType.UINT16;
		else if (type.equals("int32"))
			return ScalarType.INT32;
		else if (type.equals("uint32"))
			return ScalarType.UINT32;
		else if (type.equals("float32"))
			return ScalarType.FLOAT32;
		else if (type.equals("float64"))
			return ScalarType.FLOAT64;
		else if (type.equals("char"))
			return ScalarType.CHAR;
		else if (type.equals("uchar"))
			return ScalarType.UCHAR;
		else if (type.equals("short"))
			return ScalarType.SHORT;
		else if (type.equals("ushort"))
			return ScalarType.USHORT;
		else if (type.equals("int"))
			return ScalarType.INT;
		else if (type.equals("uint"))
			return ScalarType.UINT;
		else if (type.equals("float"))
			return ScalarType.FLOAT;
		else if (type.equals("double"))
			return ScalarType.DOUBLE;
		else
			return null;
	}
	
	private void reset() {
		elems = new Vector<Element>();
		file = null;
		elemIndex = 0;
		version = 0;
		format = null;
		texName = null;
	}
}
